import os
import json
import tempfile
import datetime
from arnparse import arnparse
from infinstor_lock import load_from_db

def my_sort_fnx(a):
    return a[0]

def union(ner_entities, dct, infinstor_time_spec):
    for nm in dct:
        if nm in ner_entities:
            full = ner_entities[nm] + dct[nm]
        else:
            full = dct[nm]
        full.sort(reverse=True, key=my_sort_fnx)
        ner_entities[nm] = full[:100]

def update_cache(infinstor_time_spec, ner_entities, s3client, bucket, prefix):
    tdir = tempfile.mkdtemp()
    tfile = os.path.join(tdir, "names.json")
    with open(tfile, 'w') as ofile:
        json.dump(ner_entities, ofile, indent=4)
    object_name = f"{prefix}/index/{infinstor_time_spec}/names.json"
    try:
        response = s3client.upload_file(tfile, bucket, object_name)
    except Exception as ex:
        print(f"Caught {ex} while uploading names entry for timespec {infinstor_time_spec}. Objectname={object_name}")

def my_filelisting_sort_fnx(a):
    return a[2]

def get_files_list_one_group(s3client, bucket, prefix, olg, resources, rv):
    print(f"get_files_list_one_group: Entered. group={olg}")
    if olg:
        prefix = prefix + olg
    print(f"get_files_list_one_group: prefix={prefix}")
    nextContinuationToken = None
    while True:
        if nextContinuationToken:
            resp = s3client.list_objects_v2(Bucket=bucket, Delimiter='/', Prefix=prefix, ContinuationToken=nextContinuationToken)
        else:
            resp = s3client.list_objects_v2(Bucket=bucket, Delimiter='/', Prefix=prefix)
        if 'Contents' in resp:
            for one in resp['Contents']:
                nm = one['Key']
                if not nm[-1] == '/':
                    if resources:
                        print(f"resources specified. evaluating file {nm} of sz {one['Size']} last_mod {one['LastModified']}", flush=True)
                        for res in resources:
                            arn = arnparse(res)
                            munged_resource_id = arn.resource.replace('/', '_')
                            if munged_resource_id in nm:
                                print(f"Add: {nm}, size {one['Size']}, last_mod {one['LastModified']}. res_id={munged_resource_id}", flush=True)
                                rv.append([nm, one['Size'], one['LastModified']])
                            else:
                                print(f"Skip: {nm}, size {one['Size']}, last_mod {one['LastModified']}. res_id={munged_resource_id}", flush=True)
                    else:
                        print(f"Add: {nm} of sz {one['Size']} last_mod {one['LastModified']}", flush=True)
                        rv.append([nm, one['Size'], one['LastModified']])
        if not resp['IsTruncated']:
            break
        else:
            nextContinuationToken = resp['NextContinuationToken']

def resources_for_tag(s3client, bucket, prefix, tag):
    import botocore
    ind = tag.find('=')
    if ind == -1:
        print('resources_for_tag: tag does not have = ???')
        return None
    tagkey = tag[:ind]
    tagval = tag[ind+1:]
    print(f"resources_for_tag: key={tagkey}, val={tagval}")

    object_name = f"{prefix}index/resource_tag_mapping_list.json"
    lfn = "/tmp/resource_tag_mapping_list.json"
    retval = []
    try:
        s3client.download_file(bucket, object_name, lfn)
        with open(lfn, 'r') as fp:
            resources = json.load(fp)
            for res in resources:
                print(f"processing resource {res['ResourceARN']}")
                for tag in res['Tags']:
                    if tag['Key'] == tagkey:
                        retval.append(res['ResourceARN'])
        os.remove(lfn)
        print(f"resources_for_tag: resources rv={retval}")
        return retval
    except botocore.exceptions.ClientError as e:
        if e.response['Error']['Code'] == "404":
            print(f"resources_for_tag: {object_name} does not exist")
            return None
        else:
            print(f"resources_for_tag: Caught {e} while reading {object_name}. Returning NotPresent")
            return None

def get_files_list(s3client, bucket, prefix, log_groups, tag):
    prefix = prefix.rstrip('/') + '/'
    rv = []
    if log_groups:
        lga = log_groups.split(',')
        for olg in lga:
            get_files_list_one_group(s3client, bucket, prefix, olg.replace('/', '_'), None, rv)
    elif tag:
        resources = resources_for_tag(s3client, bucket, prefix, tag)
        if resources:
            get_files_list_one_group(s3client, bucket, prefix, None, resources, rv)
        else:
            print(f"tag specified={tag}, but could not get resources. Ignoring tag..")
            get_files_list_one_group(s3client, bucket, prefix, None, None, rv)
    else:
        get_files_list_one_group(s3client, bucket, prefix, None, None, rv)
    rv.sort(reverse=True, key=my_filelisting_sort_fnx)
    return rv

def process_file(client, bucket, key, ner_entities, infinstor_time_spec):
    try:
        dnm = os.path.join('/tmp', key[key.rindex('/') + 1:])
        print(f"Downloading object {key} to local file {dnm}")
        client.download_file(bucket, key, dnm)
        dct = load_from_db(dnm, tag)
        union(ner_entities, dct, infinstor_time_spec)
        os.remove(dnm)
        return True
    except Exception as e:
        print(f"Caught {e} while downloading {key} from bucket {bucket}. Ignoring and trying next object..")
    return False

def populate_names(bucket, prefix, infinstor_time_spec, log_groups, tag):
    print(f'populate_names: Entered. bucket={bucket}, prefix={prefix}, infinstor_time_spec={infinstor_time_spec}, log_groups={log_groups}, tag={tag}')
    start_time = datetime.datetime.utcnow()

    import boto3
    # first list files in reverse chrono order
    try:
        s3client = boto3.client('s3', infinstor_time_spec=infinstor_time_spec)
        files = get_files_list(s3client, bucket, prefix, log_groups, tag)
    except Exception as ex:
        print(f'Caught {ex} while list_objects_v2 of {bucket} prefix {prefix} time {infinstor_time_spec}', flush=True)

    print(f"Reverse Chronological Order Files:")
    for fl in files:
        print(f"{fl[2]} : sz={fl[1]} : {fl[0]}", flush=True)

    # next, read each file and fill ner_entities
    ner_entities = {}
    total_sz = 0
    for one_entry in files:
        print(f"Processing file {one_entry[0]} last_modified {one_entry[2]}")
        if process_file(s3client, bucket, one_entry[0], ner_entities, infinstor_time_spec):
            total_sz = total_sz + one_entry[1]
        if total_sz > (512 * 1024 * 1024):
            print(f"Stopping after processing files of size {total_sz}")
            break
        else:
            print(f"Continuing after processing {total_sz} bytes")
        tnow = datetime.datetime.utcnow()
        delta = tnow - start_time
        if delta.total_seconds() > 600:
            print(f"Stopping after working for {delta.total_seconds()} seconds")
            break
        else:
            print(f"Continuing after working for {delta.total_seconds()} seconds")

    # finally, write ner_entites to s3
    update_cache(infinstor_time_spec, ner_entities, s3client, bucket, prefix)
    return ner_entities
